<?php

namespace Drp\Peek\Environment;

use Dotenv\Dotenv;
use Dotenv\Environment\Adapter\EnvConstAdapter;
use Dotenv\Environment\Adapter\PutenvAdapter;
use Dotenv\Environment\Adapter\ServerConstAdapter;
use Dotenv\Environment\DotenvFactory;
use Dotenv\Exception\ExceptionInterface;
use Drp\Peek\Environment\Exceptions\InvalidEnvVariables;
use Drp\Peek\Facade\Peek;
use Illuminate\Support\Collection;
use Illuminate\Support\Env;
use Illuminate\Support\Facades\File;
use Illuminate\Support\Facades\Storage;
use InvalidArgumentException;

class Environment
{
    public function __construct(
        string $version,
        string $environmentPath,
        string $environmentFile,
        array $bannedVariables
    ) {
        $this->appVersion = $version;
        $this->environmentPath = $environmentPath;
        $this->environmentFile = $environmentFile;
        $this->bannedVariables = new BannedVariables($bannedVariables);
        $this->dotEnv = $this->loadDotEnv();
    }

    public function variables(): Collection
    {
        return collect($this->getEnvironmentVariableNames())
            ->map(function ($envVariable) {
                if (isset($_ENV[$envVariable]) === false) {
                    return null;
                }

                $variable = new Variable($envVariable, $_ENV[$envVariable]);
                if ($this->bannedVariables->isBanned($envVariable)) {
                    return $variable->banned();
                }

                return $variable;
            })
            ->filter()
            ->sortBy(function (Variable $variable): string {
                return $variable->name();
            })
            ->keyBy(function (Variable $variable) {
                return $variable->name();
            });
    }

    public function write(array $variables): void
    {
        $this->validateNewVariables($variables);

        (new Writer())->put($variables, $this->bannedVariables);

        if (File::exists(app()->getCachedConfigPath()) === false) {
            return;
        }

        $this->loadDotEnv();
        $resource = fopen('php://memory', 'w+');
        Peek::writeCommandToStream('config:cache', [], $resource);
        fclose($resource);
    }

    /** @return \Dotenv|\Dotenv\Dotenv|null */
    protected function loadDotEnv()
    {
        return tap($this->makeDotEnv(), function (?Dotenv $dotEnv): void {
            if ($dotEnv === null || method_exists($dotEnv, 'overload') === false) {
                return;
            }

            $dotEnv->overload();
        });
    }

    private function getEnvironmentVariableNames(): array
    {
        if ($this->appVersion < '5.8') {
            $matches = [];
            preg_match_all(
                '/([A-z0-9]+)\=/m',
                File::get($this->environmentPath . '/' . $this->environmentFile),
                $matches
            );

            return $matches[1];
        }

        if ($this->appVersion >= '7.0') {
            return array_keys($this->dotEnv->load());
        }

        return $this->dotEnv->getEnvironmentVariableNames();
    }

    protected function makeDotEnv(): ?Dotenv
    {
        if ($this->appVersion < '5.6') {
            $this->makeDotEnvMutable(function (): void {
                \Dotenv::load($this->environmentPath, $this->environmentFile);
            });

            return null;
        }

        if ($this->appVersion >= '5.6' && $this->appVersion < '5.8') {
            return new Dotenv($this->environmentPath, $this->environmentFile);
        }

        if ($this->appVersion >= '7.0') {
            return Dotenv::create(
                Env::getRepository(),
                $this->environmentPath,
                $this->environmentFile
            );
        }

        return Dotenv::create(
            $this->environmentPath,
            $this->environmentFile,
            new DotenvFactory([new EnvConstAdapter(), new ServerConstAdapter(), new PutenvAdapter()])
        );
    }

    private function makeDotEnvMutable(callable $callback): void
    {
        \Dotenv::makeMutable();
        $callback();
        \Dotenv::makeImmutable();
    }

    protected function validateNewVariables(array $variables): void
    {
        Storage::disk('peek:environment')->put('.env', '');
        (new Writer('.env', 'peek:environment'))->put($variables, $this->bannedVariables);

        try {
            $environment = new Environment(
                $this->appVersion,
                Storage::disk('peek:environment')->getAdapter()->getPathPrefix(),
                '.env',
                $this->bannedVariables->toArray()
            );

            foreach ($variables as $key => $value) {
                $variable = $environment->variables()->get($key);

                if ($variable === null || $variable->value() !== (string) $value) {
                    throw InvalidEnvVariables::invalidMatch($key);
                }
            }
        } catch (ExceptionInterface | InvalidArgumentException $exception) {
            throw InvalidEnvVariables::fromPrevious($exception);
        } finally {
            Storage::disk('peek:environment')->delete('.env');
        }
    }
}
