<?php

namespace Drp\Peek;

use Carbon\Carbon;
use Drp\Peek\Environment\Environment;
use Illuminate\Contracts\Auth\Access\Gate;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Artisan;
use Illuminate\Support\Facades\Storage;
use Illuminate\Support\Facades\URL;
use Illuminate\Support\Str;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\ArrayInput;
use Symfony\Component\Console\Output\StreamOutput;

class Peek
{
    public const PEEK_ACCESS = 'peek:access';

    /** @var Gate */
    protected $gate;

    /** @var \Closure */
    protected $defaultGate;

    /** @var callable|null */
    protected $cachedCallback;

    public function __construct(Gate $gate)
    {
        $this->gate = $gate;
        $this->defaultGate = function ($user = null) { //phpcs:ignore
            return app()->environment(['local', 'development']);
        };
    }

    public function isDisabled(): bool
    {
        if (app()->version() < '5.8') {
            return call_user_func($this->cachedCallback ?? $this->defaultGate) === false;
        }

        return $this->gate->denies(self::PEEK_ACCESS);
    }

    public function gate(callable $callback): void
    {
        $this->defineGate($callback);
    }

    public function defineGate(callable $callback): void
    {
        if (app()->version() < '5.8') {
            $this->cachedCallback = $callback;

            return;
        }

        $this->gate->define(self::PEEK_ACCESS, $callback ?? $this->defaultGate);
    }

    public function enableForAll(): void
    {
        $this->defineGate(function ($user = null) { //phpcs:ignore
            return true;
        });
    }

    public function shouldMenuItemHighlight(string $url): bool
    {
        $current = str_replace('/' . config('peek.routing.root'), '', parse_url(URL::current(), PHP_URL_PATH));
        $menuUrl = str_replace('/' . config('peek.routing.root'), '', parse_url($url, PHP_URL_PATH));

        if ($current === $menuUrl) {
            return true;
        }

        if (Str::startsWith($current, $menuUrl) === false) {
            return false;
        }

        $currentParts = explode('/', trim($current, '/'));
        $menuPathParts = explode('/', trim($menuUrl, '/'));
        $diffedParts = array_diff($currentParts, $menuPathParts);

        if (count($diffedParts) === 0) {
            return true;
        }

        $diffedKey = array_keys($diffedParts)[0];
        return array_key_exists($diffedKey, $menuPathParts) === false;
    }

    public function environment(): Environment
    {
        return app(Environment::class);
    }

    public function commands(): Collection
    {
        $regexes = collect(app('config')->get('peek.commands.hidden'))->map(
            function ($hiddenCommand) {
                return str_replace('*', '.*', $hiddenCommand);
            }
        );

        $commands = collect(Artisan::all());
        $hiddenCommands = $commands->keys()->filter(function (string $commandName) use ($regexes): bool {
            foreach ($regexes as $regex) {
                if (count(preg_grep("/$regex/", [$commandName])) > 0) {
                    return true;
                }
            }

            return false;
        });

        return $this->diffKeys($commands, $hiddenCommands->flip())->sortBy(function (Command $command) {
            return $command->getName();
        });
    }

    public function commandLogs(string $commandName): Collection
    {
        $folder = Str::slug($commandName);
        $logs = Storage::disk('peek:commands')->files($folder);

        return collect($logs)->map(function ($file) {
            return (object) [
                'file' => pathinfo($file)['basename'],
                'fullPath' => $file,
                'lastModified' => Carbon::createFromTimestamp(Storage::disk('peek:commands')->lastModified($file)),
            ];
        })->sortByDesc(function ($file) {
            return $file->lastModified->timestamp;
        })->keyBy(function ($file) {
            return $file->file;
        });
    }

    /** @param resource $stream */
    public function writeCommandToStream(string $command, array $parameters, $stream): void
    {
        if (app()->version() >= '5.6') {
            Artisan::call(
                $command,
                $parameters,
                new StreamOutput($stream, StreamOutput::VERBOSITY_VERBOSE, true)
            );

            return;
        }

        $kernel = Artisan::getFacadeRoot();
        /** @var \Illuminate\Console\Application $app */
        $app = (function () {
            return $this->getArtisan();
        })->call($kernel);

        $parameters = collect($parameters)->prepend($command);
        $app->setCatchExceptions(false);

        $app->run(
            new ArrayInput($parameters->toArray()),
            new StreamOutput($stream, StreamOutput::VERBOSITY_VERBOSE, true)
        );

        $app->setCatchExceptions(true);
    }

    public function logs(): Collection
    {
        return collect(Storage::disk('peek:logs')->allFiles())
            ->filter(function (string $file) {
                return count(preg_grep('/\.log$/', [$file])) > 0;
            })
            ->keyBy(function ($file) {
                return $file;
            })
            ->map(function ($file) {
                return (object) [
                    'name' => $file,
                    'size' => number_format(Storage::disk('peek:logs')->size($file) / 1000 / 1000, 2) . ' mb',
                ];
            });
    }

    public function dashboardBlocks(): array
    {
        return array_map('app', config('peek.dashboard_blocks'));
    }

    private function diffKeys(Collection $itemsA, Collection $itemsB): Collection
    {
        return collect(array_diff_key($itemsA->all(), $itemsB->all()));
    }
}
